use itertools::Itertools;
use radix_fmt::radix;
use std::collections::HashMap;

use crate::arg;
use crate::parser;

pub trait ToCode {
    fn to_code(&self) -> String;
}

pub trait ByteResolvable<T> {
    fn resolve_number(&self, data: &mut T) -> Result<u16, String>;

    fn resolve_bytes(&self, data: &mut T) -> Result<Vec<u8>, String>;
}

#[derive(Debug)]
pub struct Data {
    pub dir: String,
    pub program: [u8; 32 * 1024],
    pub offset: u16,
    pub contants: HashMap<String, arg::Arg>,
}

impl Data {
    pub fn new(dir: String) -> Self {
        Self {
            dir,
            program: [0; 32 * 1024],
            offset: 0,
            contants: HashMap::new(),
        }
    }
}

pub fn assemble(ast: parser::ast::AST, data: &mut Data) -> Result<(), String> {
    for node in ast.body {
        data.contants
            .insert("OFFSET".to_string(), arg::Arg::Number(data.offset));

        match node {
            parser::ast::Node::Comment(_) => {}
            parser::ast::Node::Label(x) => {
                if data.contants.contains_key(&x) {
                    return Err(format!("Label already exists: '{x}'"));
                }

                data.contants.insert(x, arg::Arg::Number(data.offset));
            }
            parser::ast::Node::Call { name, arg } => {
                let arg_num = match arg {
                    Some(x) => x.resolve_number(data).unwrap(),
                    _ => 0,
                };

                data.program[data.offset as usize] = match name.as_str() {
                    "nop" => 0x00,
                    "push" => 0x01,
                    "pop" => 0x02,
                    "ts" => 0x03,
                    "tsr" => 0x04,
                    "tss" => 0x05,
                    "tlr" => 0x06,
                    "tls" => 0x07,
                    "ld" => 0x08,
                    "st" => 0x09,
                    "dbg" => 0x0a,
                    "alu" => 0x0b,
                    "at" => 0x0c,
                    "get" => 0x0d,
                    "set" => 0x0e,
                    _ => return Err(format!("Unknown opcode: '{name}'")),
                };
                if arg_num == 0 {
                    data.program[data.offset as usize] |= 0b10000000;
                }
                data.offset += 1;

                if arg_num != 0 {
                    data.program[data.offset as usize] = (arg_num >> 8) as u8;
                    data.offset += 1;

                    data.program[data.offset as usize] = arg_num as u8;
                    data.offset += 1;
                }
            }
            parser::ast::Node::MacroCall { name, args } => {
                match name.as_str() {
                    "debug" => {
                        for arg in args {
                            let bytes = arg.resolve_bytes(data).unwrap();

                            println!("{}", arg.to_code().replace('\n', "\\n"));
                            println!("  => {}", arg.resolve_number(data).unwrap());
                            println!(
                                "  => [{}]",
                                bytes
                                    .iter()
                                    .map(|n| {
                                        let num = radix(*n, 16).to_string();
                                        format!(
                                            "0x{}{}",
                                            "00".chars().take(2 - num.len()).collect::<String>(),
                                            num
                                        )
                                    })
                                    .join(", ")
                            );
                            println!(
                                "  => \"{}\"",
                                String::from_utf8(bytes).unwrap().replace('\n', "\\n")
                            );
                        }
                        println!("==========");
                    }
                    "define" => {
                        let name = match &args[0] {
                            arg::Arg::Variable(x) | arg::Arg::String(x) => x,
                            _ => {
                                return Err(
                                    "First argument of define macro needs to be a literal-like"
                                        .to_string(),
                                )
                            }
                        };

                        if data.contants.contains_key(name) {
                            return Err(format!("Constant already exists: '{name}'"));
                        }
                        data.contants.insert(name.to_string(), (&args[1]).clone());
                    }
                    "define_override" => {
                        let name = match &args[0] {
                            arg::Arg::Variable(x) | arg::Arg::String(x) => x,
                            _ => {
                                return Err(
                                    "First argument of define macro needs to be a literal-like"
                                        .to_string(),
                                )
                            }
                        };

                        data.contants.insert(name.to_string(), (&args[1]).clone());
                    }
                    "org" => {
                        data.offset = args[0].resolve_number(data).unwrap();
                    }
                    "org_add" => {
                        data.offset += args[0].resolve_number(data).unwrap();
                    }
                    "org_sub" => {
                        data.offset -= args[0].resolve_number(data).unwrap();
                    }
                    "bytes" => {
                        for arg in args {
                            for n in arg.resolve_bytes(data).unwrap() {
                                data.program[data.offset as usize] = n;
                                data.offset += 1;
                            }
                        }
                    }
                    "bw" => {
                        let string_arg = args[0].resolve_bytes(data).unwrap();
                        let string = String::from_utf8(string_arg).unwrap().replace("\\n", "\n");

                        for n in string.bytes() {
                            data.program[data.offset as usize] = n;
                            data.offset += 1;
                        }
                    }
                    _ => return Err(format!("Unknown macro: '{name}'")),
                };
            }
        }

        if data.offset > (32 * 1024) {
            return Err(format!(
                "Offset out of bounds: 0x{} > 0x8000",
                radix(data.offset, 16),
            ));
        }
    }

    Ok(())
}
