use itertools::PeekingNext;

use crate::arg;
use crate::lexer;

pub mod ast;

pub fn parse(tokens: Vec<lexer::Token>) -> Result<ast::AST, &'static str> {
    let mut iter = tokens.iter().peekable();
    let mut body: Vec<ast::Node> = Vec::new();

    while let Some(&token) = iter.peek() {
        match token {
            lexer::Token::Comment(x) => {
                body.push(ast::Node::Comment(x.trim().to_string()));
                iter.next();
            }
            lexer::Token::MacroLiteral(x) => {
                iter.next();

                body.push(ast::Node::MacroCall {
                    name: (&x[1..]).to_string(),
                    args: arg::parse_args(
                        iter.by_ref()
                            .take_while(|t| !matches!(t, lexer::Token::Newline(_)))
                            .filter(|t| !matches!(t, lexer::Token::Whitespace(_)))
                            .collect(),
                    )
                    .unwrap(),
                });
            }
            lexer::Token::Literal(x) => {
                iter.next();
                if iter
                    .peeking_next(|t| matches!(t, lexer::Token::Colon))
                    .is_some()
                {
                    body.push(ast::Node::Label(x.clone()));
                } else {
                    let args = arg::parse_args(
                        iter.by_ref()
                            .take_while(|t| !matches!(t, lexer::Token::Newline(_)))
                            .filter(|t| !matches!(t, lexer::Token::Whitespace(_)))
                            .collect(),
                    )
                    .unwrap();
                    if args.len() > 1 {
                        return Err("Opcode call only accepts one argument");
                    }

                    body.push(ast::Node::Call {
                        name: x.clone(),
                        arg: match args.first() {
                            Some(x) => Some(x.clone()),
                            _ => None,
                        },
                    });
                }
            }
            lexer::Token::Whitespace(_) | lexer::Token::Newline(_) => {
                iter.next();
            }
            _ => return Err("Unexpected token"),
        }
    }

    Ok(ast::AST { body })
}
