use hence;
use itertools::Itertools;

#[derive(Debug)]
pub enum Token {
    Comment(String),

    Newline(String),
    Whitespace(String),

    LParen,
    RParen,

    String(String),
    Number(String),
    Symbol(String),
    Literal(String),
}

impl hence::assembler::ToCode for Token {
    fn to_code(&self) -> String {
        match self {
            Token::Comment(x) => format!(";;{x}"),
            Token::Newline(x) | Token::Whitespace(x) => x.clone(),
            Token::LParen => "(".to_string(),
            Token::RParen => ")".to_string(),
            Token::String(x) => format!("\"{x}\""),
            Token::Number(x) | Token::Symbol(x) | Token::Literal(x) => x.clone(),
        }
    }
}

pub fn lex(source: String) -> Result<Vec<Token>, String> {
    let mut chars = source.chars().peekable();
    let mut tokens: Vec<Token> = Vec::new();

    while let Some(&ch) = chars.peek() {
        match ch {
            ';' => {
                chars.next();
                chars.next_if(|c| *c == ';');

                tokens.push(Token::Comment(
                    chars.peeking_take_while(|c| *c != '\n').collect(),
                ));
            }
            '\n' => {
                tokens.push(Token::Newline(
                    chars.peeking_take_while(|c| *c == '\n').collect(),
                ));
            }
            _ if ch.is_whitespace() => {
                tokens.push(Token::Whitespace(
                    chars
                        .peeking_take_while(|c| c.is_whitespace() && *c != '\n')
                        .collect(),
                ));
            }
            '(' => {
                tokens.push(Token::LParen);
                chars.next();
            }
            ')' => {
                tokens.push(Token::RParen);
                chars.next();
            }
            '"' => {
                chars.next();
                tokens.push(Token::String(
                    chars.by_ref().take_while(|c| *c != '"').collect(),
                ));
            }
            _ if ch.is_numeric() => {
                tokens.push(Token::Number(
                    chars
                        .peeking_take_while(|c| c.is_alphanumeric())
                        .collect::<String>(),
                ));
            }
            ':' => {
                chars.next();
                tokens.push(Token::Symbol(
                    chars
                        .peeking_take_while(|c| !c.is_whitespace() && *c != '(' && *c != ')')
                        .collect(),
                ));
            }
            _ => {
                tokens.push(Token::Literal(
                    chars
                        .peeking_take_while(|c| !c.is_whitespace() && *c != '(' && *c != ')')
                        .collect(),
                ));
            }
        }
    }

    Ok(tokens)
}
