use crate::lexer;

pub mod ast;

pub fn parse(tokens: Vec<lexer::Token>) -> Result<ast::AST, &'static str> {
    let mut iter = tokens.into_iter().peekable();
    let mut body: Vec<ast::Node> = Vec::new();

    while let Some(token) = iter.peek() {
        match token {
            lexer::Token::Comment(x) => {
                body.push(ast::Node::Comment(x.trim().to_string()));
                iter.next();
            }
            lexer::Token::Newline(_) | lexer::Token::Whitespace(_) => {
                iter.next();
            }
            lexer::Token::LParen => {
                iter.next();
                let mut depth: usize = 1;
                body.push(ast::Node::List(
                    parse(
                        iter.by_ref()
                            .take_while(|t| match t {
                                lexer::Token::LParen => {
                                    depth += 1;
                                    true
                                }
                                lexer::Token::RParen => {
                                    depth -= 1;
                                    depth != 0
                                }
                                _ => true,
                            })
                            .collect(),
                    )
                    .unwrap()
                    .body,
                ));
            }
            lexer::Token::RParen => return Err("Unexpected right parenthesis"),
            lexer::Token::String(x) => {
                body.push(ast::Node::String(x.clone()));
                iter.next();
            }
            lexer::Token::Number(x) => {
                body.push(ast::Node::Number(match num_parse::parse_int(x) {
                    Some(y) => y,
                    _ => return Err("Error parsing number"),
                }));
                iter.next();
            }
            lexer::Token::Symbol(x) => {
                body.push(ast::Node::Symbol(x.clone()));
                iter.next();
            }
            lexer::Token::Literal(x) => {
                body.push(ast::Node::Literal(x.clone()));
                iter.next();
            }
        }
    }

    Ok(ast::AST { body })
}
